#!/bin/env python
# -*- coding: utf-8 -*-

"""
Stephen Po-Chedley August 12, 2019

Feedback library functions

@author: pochedls
"""

import sys
sys.path.append('tropo/')
import tropo
import numpy as np
import cdms2
import MV2
import genutil

def get_kernel_map():
	"""
	kernels, kernelLookup = get_kernel_map()

	Function simply returns a dictionary of filenames and variables for
	each of the kernels ['soden', 'shell', 'pendergrass']

	"""	

	# Dictionary of filenames and variables for each kernel
	kernels = ['lwt', 'lwts', 'lwq', 'swq', 'swa', 'lwtcs', 'lwtscs', 'lwqcs', 'swqcs', 'swacs']
	kernelLookup = {'soden' : {'lwt' : ['lw_t_gfdl_std.nc', 'lw_t'],
							   'lwq' : ['lw_q_gfdl_std.nc', 'lw_q'],
							   'swq' : ['sw_q_gfdl_std.nc', 'sw_q'],
							   'swa' : ['sw_a_gfdl_std.nc', 'sw_a'],
							   'lwtcs' : ['lwclr_t_gfdl_std.nc', 'lwclr_t'],
							   'lwqcs' : ['lwclr_q_gfdl_std.nc', 'lwclr_q'],
							   'swqcs' : ['swclr_q_gfdl_std.nc', 'swclr_q'],
							   'swacs' : ['swclr_a_gfdl_std.nc', 'sw_a']},
					'shell' : {'lwt' : ['CAM3_planck_lw_kernel.nc', 'kernel_p'],
							   'lwts' : ['CAM3_surft_lw_kernel.nc', 'FLNTA_FLNT'],
							   'lwq' : ['CAM3_wv_lw_kernel.nc', 'kernel_p'],
							   'swq' : ['CAM3_wv_sw_kernel.nc', 'kernel_p'],
							   'swa' : ['CAM3_albedo_sw_kernel.nc', 'monkernel'],
							   'lwtcs' : ['CAM3_planck_lw_clr_kernel.nc', 'kernel_p'],
							   'lwtscs' : ['CAM3_surft_lw_clr_kernel.nc', 'FLNTCA_FLNTC'],
							   'lwqcs' : ['CAM3_wv_lw_clr_kernel.nc', 'kernel_p'],
							   'swqcs' : ['CAM3_wv_sw_clr_kernel.nc', 'kernel_p'],
							   'swacs' : ['CAM3_albedo_sw_clr_kernel.nc', 'monkernel']},
					'pendergrass' : {'lwt' : ['t.kernel.plev.nc', 'FLNT'],
							   'lwts' : ['ts.kernel.nc', 'FLNT'],
							   'lwq' : ['q.kernel.plev.nc', 'FLNT'],
							   'swq' : ['q.kernel.plev.nc', 'FSNT'],
							   'swa' : ['alb.kernel.nc', 'FSNT'],
							   'lwtcs' : ['t.kernel.plev.nc', 'FLNTC'],
							   'lwtscs' : ['ts.kernel.nc', 'FLNTC'],
							   'lwqcs' : ['q.kernel.plev.nc', 'FLNTC'],
							   'swqcs' : ['q.kernel.plev.nc', 'FSNTC'],
							   'swacs' : ['alb.kernel.nc', 'FSNTC']}								   				   
							   }
	return kernels, kernelLookup

def load_kernels(kernelSource, kernelPath='kernels/'):
	"""
	lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs = load_kernels(kernelSource, kernelPath='kernels/')

	Function loads and returns the kernels (cs denoted clear-sky):
			lwt 		longwave temperature
			lwts 		longwave surface temperature
			lwq 		longwave water vapor
			swq 		shortwave water vapor
			swa 		shortwave albedo
	
	with axes [month, [pressure], lat, lon] for a given kernelSource ('soden', 'shell', 'pendergrass')

	"""		
	kernels, kernelLookup = get_kernel_map()
	# loop over kernels and load fields
	d = locals()
	for kernel in kernels:
		# if kernel doesn't exist, continue... 
		if kernel not in kernelLookup[kernelSource]:
			exec(kernel + ' = []', globals(), locals())
			continue;
		# construct filename
		fn = kernelPath + '/' + kernelSource + '/' + kernelLookup[kernelSource][kernel][0]
		vid = kernelLookup[kernelSource][kernel][1]
		# load data
		f = cdms2.open(fn)
		data = f(vid)
		# attach pressure axis if Pendergrass
		if kernelSource == 'pendergrass':
			if len(data.shape) == 4:
				plev = f('plev')
				plev = cdms2.createAxis(plev)
				plev.id = 'plev'
				plev.positive = 'down'
				plev.units = 'Pa'
				plev.axis = 'Z'
				data.setAxis(1, plev)
		# recast as generic kernel name
		exec(kernel + ' = data.copy()', globals(), locals())
		# exec(kernel + ' = data.copy()')
		f.close()

	return d['lwt'], d['lwts'], d['lwq'], d['swq'], d['swa'], d['lwtcs'], d['lwtscs'], d['lwqcs'], d['swqcs'], d['swacs']

def standardize_kernels(kernelSource, lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs, kernelPath='kernels/'):
	"""
	lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs 
			= standardize_kernels(kernelSource, lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs, kernelPath='kernels/')

	Function standardizes the kernels so they all have the same form and units from a given kernelSource
	('soden', 'shell', 'pendergrass'). 

	"""		
	# create standard plev axis
	plevAxis = np.array([100000.,  92500.,  85000.,  70000.,  60000.,  50000.,  40000., 30000.,  25000.,  20000.,  15000.,  10000.,   7000.,   5000., 3000.,   2000.,   1000.])
	plevAxis = (plevAxis[0:-1] + plevAxis[1:]) / 2.
	plevAxis = cdms2.createAxis(plevAxis)
	plevAxis.id = 'plev'
	plevAxis.units = 'hPa'
	# If Soden or Shell reverse layer ordering
	if kernelSource in ['soden', 'shell']:
		lwt = lwt[:, ::-1, :, :]
		lwq = lwq[:, ::-1, :, :]
		swq = swq[:, ::-1, :, :]
		lwtcs = lwtcs[:, ::-1, :, :]
		lwqcs = lwqcs[:, ::-1, :, :]
		swqcs = swqcs[:, ::-1, :, :]
	# Convert to albedo change (rather than percent change)
	swa = swa * 100.
	swacs = swacs * 100.
	# If Soden get surface field from 3d field
	if kernelSource == 'soden':
		# get surface kernels
		lwts = lwt[:, 0, : , :]
		lwtscs = lwtcs[:, 0, : , :]
		# take surface out of 3d kernel
		lwt = lwt[:, 1:, :, :]
		lwtcs = lwtcs[:, 1:, :, :]
		lwq = lwq[:, 1:, :, :]
		swq = swq[:, 1:, :, :]
		lwqcs = lwqcs[:, 1:, :, :]
		swqcs = swqcs[:, 1:, :, :]
		# lose singleton dimensions
		swa = swa[:, 0, :, :]
		swacs = swacs[:, 0, :, :]
	# if Pendergrass, fill in values below surface (with values from above),
	# add zeros above 100 hPa, and convert to per 100 hPa
	if kernelSource == 'pendergrass':
		# fill values below surface (with value from layer above)
		# this allows us to calculate layer averages below...
		lwt = fill_below_surface(lwt)
		lwtcs = fill_below_surface(lwtcs)
		lwq = fill_below_surface(lwq)
		lwqcs = fill_below_surface(lwqcs)
		swq = fill_below_surface(swq)
		swqcs = fill_below_surface(swqcs)
		lat = lwt.getLatitude()
		lon = lwt.getLongitude()
		# append on zeros for stratosphere
		zeroMatrix = np.zeros((12, 5, lwt.shape[2], lwt.shape[3]))
		lwt = MV2.concatenate((lwt, zeroMatrix),1)
		lwtcs = MV2.concatenate((lwtcs, zeroMatrix),1)
		lwq = MV2.concatenate((lwq, zeroMatrix),1)
		lwqcs = MV2.concatenate((lwqcs, zeroMatrix),1)
		swq = MV2.concatenate((swq, zeroMatrix),1)
		swqcs = MV2.concatenate((swqcs, zeroMatrix),1)
		lwt.setAxis(2, lat)
		lwt.setAxis(3, lon)
		lwtcs.setAxis(2, lat)
		lwtcs.setAxis(3, lon)
		lwq.setAxis(2, lat)
		lwq.setAxis(3, lon)
		lwqcs.setAxis(2, lat)
		lwqcs.setAxis(3, lon)
		swq.setAxis(2, lat)
		swq.setAxis(3, lon)
		swqcs.setAxis(2, lat)
		swqcs.setAxis(3, lon)
		lwt = lwt * 100.
		lwtcs = lwtcs * 100.
		lwq = lwq * 100.
		lwqcs = lwqcs * 100.
		swq = swq * 100.
		swqcs = swqcs * 100.

	# get layer values for 3d kernels	
	lwt = (lwt[:, 0:-1, :, :] + lwt[:, 1:, :, :]) / 2.
	lwtcs = (lwtcs[:, 0:-1, :, :] + lwtcs[:, 1:, :, :]) / 2.

	lwt.setAxis(1, plevAxis)
	lwtcs.setAxis(1, plevAxis)

	return lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs

def fill_below_surface(data):
	"""
	dataOut = fill_below_surface(data)

	Function fills lower layers with data from the layer above. Assumes the near 
	surface is the first pressure level. 

	"""		
	for i in range(data.shape[1] - 1):
		fieldAbove = data[:, -i-1, :, :]
		fieldBelow = data[:, -i-2, :, :]
		data[:, -i-2, :, :] = MV2.where(fieldBelow.mask, fieldAbove, fieldBelow)
	return data

def dlogq_kernel(ta_ctrl, kernel):

	# get the models dlog(q)/dT sensitivity
	e = calc_qs(ta_ctrl) 
	qs = e * 0.622 / (np.reshape(ta_ctrl.getLevel()[:],[1,17,1,1]) - e)
	e1 = calc_qs(ta_ctrl + 1) 
	qs1 = e1 * 0.622 / (np.reshape(ta_ctrl.getLevel()[:],[1,17,1,1]) - e1)
	s = np.log(qs1/qs)
	kernel['lwqNorm'] = kernel['lwq'] / s
	kernel['swqNorm'] = kernel['swq'] / s
	kernel['lwqcsNorm'] = kernel['lwqcs'] / s
	kernel['swqcsNorm'] = kernel['swqcs'] / s

	# get layer values
	kernel['lwqNorm'] = (kernel['lwqNorm'][:, 0:-1, :, :] + kernel['lwqNorm'][:, 1:, :, :]) / 2.
	kernel['swqNorm'] = (kernel['swqNorm'][:, 0:-1, :, :] + kernel['swqNorm'][:, 1:, :, :]) / 2.	
	kernel['lwqcsNorm'] = (kernel['lwqcsNorm'][:, 0:-1, :, :] + kernel['lwqcsNorm'][:, 1:, :, :]) / 2.
	kernel['swqcsNorm'] = (kernel['swqcsNorm'][:, 0:-1, :, :] + kernel['swqcsNorm'][:, 1:, :, :]) / 2.	

	# set axis
	plevAxis = kernel['lwt'].getLevel()
	kernel['lwqNorm'].setAxis(1, plevAxis)
	kernel['swqNorm'].setAxis(1, plevAxis)	
	kernel['lwqcsNorm'].setAxis(1, plevAxis)
	kernel['swqcsNorm'].setAxis(1, plevAxis)	

	return kernel


def get_standard_kernels(kernelSource, kernelPath='kernels/'):
	"""
	kernels = get_standard_kernels(kernelSource, kernelPath='kernels/')

	Function returns standardized kernels for a given kernelSource ('soden', 'shell', 'pendergrass'). 

	"""		
	# load kernels
	lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs = load_kernels(kernelSource, kernelPath=kernelPath)
	# standardize kernels
	lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs = standardize_kernels(kernelSource, lwt, lwts, lwq, swq, swa, lwtcs, lwtscs, lwqcs, swqcs, swacs, kernelPath=kernelPath)
	# put kernels in a dictionary
	kernels = {'lwt' : lwt, 
			   'lwts' : lwts, 
			   'lwq' : lwq, 
			   'swq' : swq, 
			   'swa' : swa, 
			   'lwtcs' : lwtcs, 
			   'lwtscs' : lwtscs, 
			   'lwqcs' : lwqcs, 
			   'swqcs' : swqcs, 
			   'swacs' : swacs}
	return kernels

def get_tropopause(ta):
	""" trop = get_tropopause(ta)

	function returns the tropopause height (Pa, numpy array) for a given
	ta field (cdms transient variable as input).

	+ tropo code from : http://www.inscc.utah.edu/~reichler/research/projects/TROPO/tropocode.htm
	+ f2py -c -m tropo tropo.f90
	"""
	plevAxis = ta.getLevel()
	trop = np.zeros((ta.shape[0], ta.shape[2], ta.shape[3])) * np.nan
	for i in range(ta.shape[0]):
		tai = np.transpose(ta[i], [2, 1, 0])
		plimu=45000 # upper limit for tropopause pressure in Pa, usually 45000
		pliml=7500 # lower limit for tropopause pressure in Pa, usually 7500
		plimlex=7500 # lower limit in extratropics, usually same as pliml, i.e., 7500
		dofill=1 # 1 = fill undefined values with neighboring points if 1, 0 = do not fill
		tp,tperr = tropo.tropo(tai,plevAxis[:]/100.,plimu,pliml,plimlex,dofill)
		trop[i] = np.transpose(tp, (1,0))
	time = ta.getTime()
	lat = ta.getLatitude()
	lon = ta.getLongitude()
	trop = cdms2.createVariable(trop)
	trop.id = 'trop'
	trop.long_name = 'tropopause height'
	trop.units = 'Pa'
	trop.setAxis(0, time)
	trop.setAxis(1, lat)
	trop.setAxis(2, lon)
	return trop

def get_weights(ps, trop, field):
	"""
	wts = get_weights(ps, trop, field)

	Function returns the layer weights (layer thickness) given the upper
	integration boundary (e.g., the tropopause, trop) and the lower bound
	(e.g., surface pressure, ps). Uses the input field to get pressure level
	information. 
		ps - response field [time, lat, lon]
		trop - radiative kernel [time, lat, lon]
		field - field that will be integrated [time, plev, lat, lon]

	"""	
	PAL = MV2.zeros(ps.shape)
	PAT = MV2.ones(ps.shape)	
	wts = MV2.zeros(field.shape - MV2.array([0,1,0,0]))
	plev = field.getLevel()[:]
	if plev[0] < plev[1]:
		raise ValueError('This script assumes that plev[0] > plev[1].')
	for i in range(len(plev)-1):
		# allocate slice 
		# sfield = np.zeros(ps.shape)
		sfield = MV2.zeros(ps.shape)
		# get first level above surface
		p = plev[i]
		pp = plev[i+1]
		ISURF = MV2.greater(ps, p)
		II = MV2.equal(PAL, 0)
		IA = MV2.logical_and(ISURF, II)		
		PAL = MV2.where(IA, pp, PAL)
		# get first layer below tropopause
		ITROP = MV2.less(trop, p)
		II = MV2.greater_equal(trop, pp)
		IB = MV2.logical_and(ITROP, II)
		PAT = MV2.where(IB, p, PAT)
		# layers above tropopause or below surface (i.e., zero weight)
		IC = MV2.logical_or(MV2.logical_not(ISURF),MV2.logical_not(ITROP))		
		# layers adjacent to both tropopause and surface (i.e., layer is between tropopause and surface)
		ID = MV2.logical_and(IA, IB)
		# layers not adjacent to tropopause or surface (i.e., full layers)
		IE = MV2.logical_or(IA, IB)
		IE = MV2.logical_and(MV2.logical_not(IC), MV2.logical_not(IE))	
		# Store values
		sfield = MV2.where(IA, ps-PAL, sfield) 
		sfield = MV2.where(IB, PAT-trop, sfield)
		sfield = MV2.where(IC, 0., sfield)
		sfield = MV2.where(ID, ps-trop, sfield)
		sfield = MV2.where(IE, p - pp, sfield)
		# store field and weight by per 100 hPa (1/100 for Pa to hPa and 1/100 for per *100* hPa)
		wts[:, i, :, :] = sfield / 10000.
	return wts

def get_layer_values(ps, field, surface_field):
	"""
	field_layer = get_layer_values(ps, field, surface_field)

	Function returns the layer values for a field on levels (using simple
	average of levels). 

	Inputs: 
			ps - surface pressure values [time, lat, lon]
			field - input field on levels [time, plev, lat, lon]
			surface_field - surface values corresponding to input field [time, lat, lon]
	Returns:
			field_layer - layer values of input field 


	https://en.wikipedia.org/wiki/Potential_temperature

	"""			
	# get pressure levels
	plev = field.getLevel()
	play = (plev[1:] + plev[0:-1]) / 2.
	play = cdms2.createAxis(play)
	play.id = 'play'
	play.units = plev.units
	play.realtopology = plev.realtopology
	play.axis = plev.axis
	play.positive = plev.positive
	play.standard_name = plev.standard_name
	# pre-allocate field
	field_layer = field.copy()
	if plev[0] < plev[1]:
		raise ValueError('This script assumes that plev[0] > plev[1].')	
	for i in range(len(plev)-1):
		# get layer i and i+1
		# field_below = field[:, i, :, :]
		# field_above = field[:, i+1, :, :]
		pbelow = plev[i] * MV2.ones(ps.shape)
		pabove = plev[i+1] * MV2.ones(ps.shape)
		# determine points in which surface is 
		# in between layers
		I = MV2.greater(ps, pabove)
		II = MV2.less_equal(ps, pbelow)
		IA = MV2.logical_and(I, II)
		# when the surface is between atmospheric pressure levels
		# set the lower level to the surface values
		field_layer[:, i, :, :]  = MV2.where(IA, surface_field, field[:, i, :, :])
	# calculate layer temperature as simple mean
	field_layer = MV2.add(field_layer[:, 0:-1, :, :], field_layer[:, 1:, :, :]) / 2.
	field_layer.setAxis(1, play)

	return field_layer

def get_3d_feedback(dx_layers, kernel_layers, wts, surface_kernel, dx_surface, idValue):
	"""
	rad = get_3d_feedback(dx_layers, kernel_layers, wts, surface_kernel, dx_surface, idValue)

	Function to convolve 3d response field (layer averages) with radiative kernels (layer
	averages) assuming trapezoidal integration. Returns the radiative response. 
		dx_layers - response field [time, plev, lat, lon]
		kernel_layers - radiative kernel [time, plev, lat, lon]
		wts - layer weights [time, plev, lat, lon]
		surface_kernel - surface kernel [time, lat, lon]
		dx_surface - surface response [time, lat, lon]
		idValue - id value for the output field

	"""
	time = dx_surface.getTime()
	lat = dx_surface.getLatitude()
	lon = dx_surface.getLongitude()
	# put zeros where temperature or kernels are masked
	wts[dx_layers.mask] = 0.
	kernel_layers[dx_layers.mask] = 0.
	dx_layers[dx_layers.mask] = 0.
	wts[kernel_layers.mask] = 0.
	dx_layers[kernel_layers.mask] = 0.	
	kernel_layers[kernel_layers.mask] = 0.
	dx_layers[wts.mask] = 0.	
	kernel_layers[wts.mask] = 0.
	wts[wts.mask] = 0.
	wts.mask=False
	dx_layers.mask=False
	kernel_layers.mask=False
	# convolve matrices
	rad = MV2.sum(MV2.multiply(MV2.multiply(dx_layers, kernel_layers), wts), axis=1) + surface_kernel * dx_surface
	rad.setAxis(0, time)
	rad.setAxis(1, lat)
	rad.setAxis(2, lon)
	rad.id = idValue
	rad.units = 'W/m**-2'
	return rad

def get_2d_feedback(surface_kernel, dx_surface, idValue):
	"""
	rad = get_2d_feedback(surface_kernel, dx_surface, idValue)

	Function to convolve 2d response field with surface radiative kernel. 
	Returns the radiative response. 
		surface_kernel - surface kernel [time, lat, lon]
		dx_surface - surface response [time, lat, lon]
		idValue - id value for the output field

	"""	
	# convolve matrices
	rad = MV2.multiply(surface_kernel, dx_surface)
	rad.id = idValue
	rad.units = 'W/m**-2'
	return rad

def calc_qs(ta):
	a1l = 611.21
	a3l = 17.502
	a4l = 32.19
	a1i = 611.21
	a3i = 22.587
	a4i = -0.7
	to = 273.16
	ti = 250.16
	el = a1l*np.exp(a3l*(ta-to)/(ta-a4l))
	ei = a1i*np.exp(a3i*(ta-to)/(ta-a4i))
	es = np.zeros(ta.shape)*np.nan
	alpha = np.power((ta-ti)/(to-ti),2)
	alpha[ta<=ti] = 0
	alpha[ta>=to] = 1
	es = el*alpha + ei*(1-alpha)
	return es

def get_feedbacks(ta_ctrl, hus_ctrl, ps_ctrl, dta, dts, dhus, dalbedo, dswcre, dlwcre, kernels):

	# remap fields to kernel grids
	gridOut = kernels['swa'].getGrid()
	ta_ctrl = ta_ctrl.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	hus_ctrl = hus_ctrl.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	ps_ctrl = ps_ctrl.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	dta = dta.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	dts = dts.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	dalbedo = dalbedo.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	dhus = dhus.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	dswcre = dswcre.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)
	dlwcre = dlwcre.regrid(gridOut, regridTool='esmf', regridMethod = 'linear', missing=np.nan, coordSys='deg', diag = {}, periodicity = 1)

	# update water vapor kernels with log-scaling
	kernels = dlogq_kernel(ta_ctrl, kernels)

	# get tropopause height (based on ta climatology)
	trop = get_tropopause(ta_ctrl)
	# get layer values
	dt_layers = get_layer_values(ps_ctrl, dta, dts)
	dlogq = MV2.subtract(np.log(hus_ctrl + dhus), np.log(hus_ctrl))
	dlogq.setAxis(1, hus_ctrl.getLevel())
	dhus_layers = get_layer_values(ps_ctrl, dlogq, np.zeros(dts.shape)) 
	# get pressure weights
	wts = get_weights(ps_ctrl, trop, dta)
	# get matrix of planck temperature change values 
	# e.g., surface_dt invariant with height but same size as ta
	dt_planck, y = genutil.grower(dts, dt_layers)
	dt_planck = dt_planck.reorder('tzyx')

	# convolve kernels, weights, and dt via trapezoidal integration
	zeroSurface = kernels['lwts'] * 0.
	# longwave feedbacks
	lapse = get_3d_feedback(dt_layers - dt_planck, -kernels['lwt'], wts, zeroSurface, zeroSurface, 'lr')
	lapseClr = get_3d_feedback(dt_layers - dt_planck, -kernels['lwtcs'], wts, zeroSurface, zeroSurface, 'lr_clr')
	planck = get_3d_feedback(dt_planck, -kernels['lwt'], wts, -kernels['lwts'], dts, 'planck')
	planckClr = get_3d_feedback(dt_planck, -kernels['lwtcs'], wts, -kernels['lwtscs'], dts, 'planck_clr')
	lwWaterVapor = get_3d_feedback(dhus_layers, -kernels['lwqNorm'], wts, zeroSurface, zeroSurface, 'lwwv')
	lwWaterVaporClr = get_3d_feedback(dhus_layers, -kernels['lwqcsNorm'], wts, zeroSurface, zeroSurface, 'lwwv_clr')
	# shortwave feedbacks
	swWaterVapor = get_3d_feedback(dhus_layers, kernels['swqNorm'], wts, zeroSurface, zeroSurface, 'swwv')
	swWaterVaporClr = get_3d_feedback(dhus_layers, kernels['swqcsNorm'], wts, zeroSurface, zeroSurface, 'swwv_clr')	
	# total water vapor feedback
	waterVapor = MV2.add(lwWaterVapor, swWaterVapor)
	# get albedo feedbacks
	albedo = get_2d_feedback(kernels['swa'], dalbedo, 'albedo')
	albedoClr = get_2d_feedback(kernels['swacs'], dalbedo, 'albedo_clr')
	# cloud feedbacks
	lwMask = (lapse - lapseClr) + (planck - planckClr) + (lwWaterVapor - lwWaterVaporClr)
	swMask = (swWaterVapor - swWaterVaporClr) + (albedo - albedoClr)
	swCloud = dswcre - swMask
	lwCloud = dlwcre - lwMask

	# take the field time units
	time = dts.getTime()
	planck.setAxis(0, time)
	lapse.setAxis(0, time)
	waterVapor.setAxis(0, time)
	albedo.setAxis(0, time)
	swCloud.setAxis(0, time)
	lwCloud.setAxis(0, time)

	feedbacks = {}
	feedbacks['planck'] = planck
	feedbacks['lapse'] = lapse
	feedbacks['waterVapor'] = waterVapor
	feedbacks['albedo'] = albedo
	feedbacks['swCloud'] = swCloud
	feedbacks['lwCloud'] = lwCloud

	return feedbacks






